import argparse
import os
import sys
from copy import deepcopy
from pathlib import Path
from typing import Any

import yaml

sys.path.append(str(Path(__file__).parent.parent.absolute()))

from naapc import nconfig, ndict


class testNDict:
    def __init__(self, delimiter=";"):
        self.nd = ndict({"this": "is test"})
        print("Pass copy test.")

        with open("test/config.yaml", "r") as f:
            self.rawd = yaml.safe_load(f)
        self.nd = ndict(self.rawd)
        self.test_init()
        print("Pass init test.")

        self.test_from_flatten_dict()
        print("Pass from flatten test.")

        self.test_paths()
        print("Pass paths test.")

        self.test_contains()
        print("Pass contains test.")

        self.test_eq()
        print("Pass eq test.")

        self.test_del()
        print("Pass del test.")

        self.test_getitem()
        print("Pass getitem test.")

        self.test_flatten()
        print("Pass flatten test.")

        self.test_setitem()
        print("Pass setitem test.")

        self.test_update()
        print("Pass update test.")

        self.test_compare()
        print("Pass compare test.")

        self.test_delimiter()
        print("Pass delimiter test.")

    def test_init(self, delimiter=";"):
        nd1 = ndict(self.nd, delimiter=delimiter)
        assert nd1 == self.nd
        assert self.nd.raw_dict is not nd1.raw_dict
        assert self.nd is not nd1

        nd1 = ndict(deepcopy(self.nd.raw_dict), delimiter=delimiter)
        assert nd1 is not self.nd
        assert nd1 == self.nd
        del nd1[f"task{delimiter}task"]
        self.nd[f"task{delimiter}task"]
        assert nd1 != self.nd

    def test_from_flatten_dict(self, delimiter=";"):
        nd1 = ndict.from_flatten_dict(self.nd, delimiter=delimiter)
        assert nd1 is not self.nd
        assert nd1._d is not self.nd._d
        assert nd1 == self.nd

        nd1 = ndict.from_flatten_dict(self.nd.flatten_dict, delimiter=delimiter)
        assert nd1 is not self.nd
        assert nd1._d is not self.nd._d
        assert nd1 == self.nd

    def test_from_list(self, delimiter=";"):
        nd1 = ndict.from_list_of_dict(self.nd.flatten_dict_split, delimiter=delimiter)
        assert nd1 == self.nd

    def test_eq(self, delimiter=";"):
        nd = deepcopy(self.nd)
        assert nd == self.nd
        nd[f"task{delimiter}task"] = "not_exist"
        assert nd != self.nd
        nd = deepcopy(self.nd)
        nd["not_exist"] = "not_exist"
        assert nd != self.nd

        d = deepcopy(self.nd.raw_dict)
        nd = ndict(d, delimiter=delimiter)
        assert nd == self.nd
        del d["task"]
        nd = ndict(d, delimiter=delimiter)
        assert nd != self.nd

    def test_paths(self, delimiter=";"):
        for p in self.nd.paths:
            self.nd[p]

        all_paths = []
        for path in self.nd.flatten_dict.keys():
            nodes = path.split(delimiter)
            all_paths.extend(delimiter.join(nodes[:i]) for i in range(1, len(nodes) + 1))
        assert set(self.nd.paths) == set(all_paths)

    def test_contains(self, delimiter=";"):
        for p in self.nd.paths:
            assert p in self.nd
            assert f"{p}{delimiter}fwejkl" not in self.nd

    def test_del(self, delimiter=";"):
        for path in self.nd.paths:
            nd = deepcopy(self.nd)
            del nd[path]
            diff = set(self.nd.paths) - set(nd.paths)
            assert all(p.startswith(path) for p in diff)
            assert nd != self.nd

    def test_getitem(self, delimiter=";"):
        def _check(pks, k, v, nd):
            path = delimiter.join(pks + [k])
            assert nd[path] == v, f"Wong at {path}. raw: {v}, nd: {nd}"
            if isinstance(v, dict):
                for nk, nv in v.items():
                    _check(pks + [k], nk, nv, nd)

        for k, v in self.rawd.items():
            _check([], k, v, self.nd)

    def test_flatten(self, delimiter=";"):
        for path, v in self.nd.flatten_dict.items():
            assert self.nd[path] == v, f"Wrong at {path}. nd: {self.nd[path]}, flatten: {v}"

        def _check(pks, k, v, flatten):
            path = delimiter.join(pks + [k])
            if isinstance(v, dict):
                for nk, nv in v.items():
                    _check(pks + [k], nk, nv, flatten)
            else:
                assert flatten[path] == v, f"Wrong at {path}. raw: {v}, flatten: {flatten[path]}"

        for k, v in self.rawd.items():
            _check([], k, v, self.nd.flatten_dict)

    def test_setitem(self, delimiter=";"):
        nd = deepcopy(self.nd)
        for path in nd.flatten_dict.keys():
            nd[path] = 1

        assert nd != self.nd
        assert set(nd.paths) == set(self.nd.paths)
        for v in nd.flatten_dict.values():
            assert v == 1

        nd = deepcopy(self.nd)
        nd[f"not_exist{delimiter}not_exist{delimiter}not_exist{delimiter}not_exist"] = "not_exist"
        assert "not_exist" in nd
        assert f"not_exist{delimiter}not_exist" in nd
        assert f"not_exist{delimiter}not_exist{delimiter}not_exist" in nd
        assert f"not_exist{delimiter}not_exist{delimiter}not_exist{delimiter}not_exist" in nd
        assert (
            nd[f"not_exist{delimiter}not_exist{delimiter}not_exist{delimiter}not_exist"]
            == "not_exist"
        )

        assert nd != self.nd

    def test_update(self, delimiter=";"):
        nd = deepcopy(self.nd)

        fd = {path: 1 for path in self.nd.flatten_dict.keys()}
        nd.update(fd)

        assert nd != self.nd
        assert set(nd.paths) == set(self.nd.paths)
        for v in nd.flatten_dict.values():
            assert v == 1

        nd1 = deepcopy(self.nd)
        nd1.update(nd)
        assert nd1 != self.nd
        assert set(nd1.paths) == set(self.nd.paths)
        for v in nd1.flatten_dict.values():
            assert v == 1

        nd = deepcopy(self.nd)
        fd = {f"not_exist{delimiter}not_exist{delimiter}not_exist": "not_exist"}
        try:
            nd.update(fd)
            raise RuntimeError
        except Exception:
            pass
        nd.update(fd, ignore_missing_path=True)
        assert "not_exist" in nd
        assert f"not_exist{delimiter}not_exist" in nd
        assert f"not_exist{delimiter}not_exist{delimiter}not_exist" in nd
        assert nd[f"not_exist{delimiter}not_exist{delimiter}not_exist"] == "not_exist"

        assert nd != self.nd

    def test_compare(self, delimiter=";"):
        nd = deepcopy(self.nd)
        assert not nd.compare_dict(self.nd)
        assert not self.nd.compare_dict(nd)
        assert nd.compare_dict(self.nd) == self.nd.compare_dict(nd)

        diff1 = {}
        diff2 = {}
        for path, v in self.nd.flatten_dict.items():
            nd[path] = "not_exist"
            diff1[path] = ("not_exist", v)
            diff2[path] = (v, "not_exist")
            assert nd.compare_dict(self.nd) == diff1
            assert self.nd.compare_dict(nd) == diff2

        nd = deepcopy(self.nd)
        diff1 = {}
        diff2 = {}
        for path, v in self.nd.flatten_dict.items():
            if v is None:
                continue
            del nd[path]
            diff1[path] = (None, v)
            diff2[path] = (v, None)
            assert nd.compare_dict(self.nd) == diff1
            assert self.nd.compare_dict(nd) == diff2

        nd = deepcopy(self.nd)
        nd[f"not_exist{delimiter}not_exist{delimiter}not_exist"] = "not_exist"
        assert nd.compare_dict(self.nd) == {
            f"not_exist{delimiter}not_exist{delimiter}not_exist": ("not_exist", None)
        }
        assert self.nd.compare_dict(nd) == {
            f"not_exist{delimiter}not_exist{delimiter}not_exist": (None, "not_exist")
        }

    def test_delimiter(self):
        delimiter = "."
        self.nd = ndict(self.rawd, delimiter=delimiter)
        self.test_init(delimiter=delimiter)
        self.test_from_flatten_dict(delimiter=delimiter)
        self.test_paths(delimiter=delimiter)
        self.test_contains(delimiter=delimiter)
        self.test_eq(delimiter=delimiter)
        self.test_del(delimiter=delimiter)
        self.test_getitem(delimiter=delimiter)
        self.test_flatten(delimiter=delimiter)
        self.test_setitem(delimiter=delimiter)
        self.test_update(delimiter=delimiter)
        self.test_compare(delimiter=delimiter)


class testNConfig:
    def __init__(self, delimiter=";"):
        with open("test/config.yaml", "r") as f:
            self.rawd = yaml.safe_load(f)
        self.nd = ndict(self.rawd, delimiter=delimiter)

        self.test_init()
        print("Pass init test.")

        self.test_parse()
        print("Pass arg parsing test.")

        with open("test/config_arg_spe.yaml", "r") as f:
            rawd_spe = yaml.safe_load(f)
        nd_spe = ndict(rawd_spe, delimiter=delimiter)

        self.config = nconfig(nd_spe, delimiter=delimiter)

        self.test_update()
        print("Pass update test.")

        self.test_setitem()
        print("Pass setitem test.")

        self.test_save()
        print("Pass save test.")

        self.test_delimiter()
        print("Pass delimiter test.")

    def test_init(self, delimiter=";"):
        with open("test/config_arg_spe.yaml", "r") as f:
            rawd_spe = yaml.safe_load(f)
        nd_spe = ndict(rawd_spe, delimiter=delimiter)

        config = nconfig(self.rawd, delimiter=delimiter)
        assert config == self.nd
        assert config != nd_spe

        config = nconfig(nd_spe, delimiter=delimiter)
        assert config._arg_specification
        assert config._d is not self.rawd

    def test_parse(self, delimiter=";"):
        def _get_test_value(v):
            def _get(v):
                if isinstance(v, bool):
                    v = 0 if v else 1
                elif isinstance(v, float):
                    v = -1000
                elif isinstance(v, int):
                    v = -100
                elif isinstance(v, str):
                    v = "no_exist"
                elif isinstance(v, type(None)):
                    v = 1
                else:
                    raise ValueError(f"Unexpected type {type(v)}")
                return v

            if isinstance(v, list):
                _v = []
                for x in v:
                    _v.append(_get(x))
                return _v
            else:
                return _get(v)

        with open("test/config_arg_spe.yaml", "r") as f:
            rawd_spe = yaml.safe_load(f)
        nd_spe = ndict(rawd_spe, delimiter=delimiter)
        config = nconfig(nd_spe, delimiter=delimiter)
        for path, v in config.flatten_dict.items():
            config = nconfig(nd_spe, delimiter=delimiter)
            parser = argparse.ArgumentParser()
            if config._arg_specification.get(path, None) == config._ignore_key:
                parser = config.add_to_argparse(parser)
                flag = f"--{path.replace(delimiter, '__')}"
                _args = [flag, "1"]
                args, extra_args = parser.parse_known_args(_args)
                assert extra_args == _args
                continue
            if path in config._arg_specification and "flag" in config._arg_specification[path]:
                flag = config._arg_specification[path]["flag"]
            else:
                flag = f"--{path.replace(delimiter, '__')}"
            if path == f"data{delimiter}dataset":
                args, extra_args = parser.parse_known_args([flag, "imagenet"])
            elif isinstance(v, list):
                args, extra_args = parser.parse_known_args(
                    [flag, *[str(x) for x in _get_test_value(v)]]
                )
            else:
                args, extra_args = parser.parse_known_args([flag, str(_get_test_value(v))])
            parser = config.add_to_argparse(parser)
            extra_args = config.parse_update(parser, extra_args)
            assert len(extra_args) == 0
            if path == f"data{delimiter}dataset":
                assert config[path] == "imagenet"
            else:
                if isinstance(config[path], bool):
                    target = bool(_get_test_value(v))
                else:
                    target = _get_test_value(v)
                assert config[path] == target
            _nd_spe = deepcopy(nd_spe)
            del _nd_spe["_ARGUMENT_SPECIFICATION"]
            for _path, _v in _nd_spe.flatten_dict.items():
                if _path != path:
                    assert config[_path] == _v
                else:
                    assert config[_path] != _v

    def test_update(self, delimiter=";"):
        config = deepcopy(self.config)

        fd = {path: 1 for path in self.config.flatten_dict.keys()}
        config.update(fd)

        assert config != self.config
        assert set(config.paths) == set(self.config.paths)
        for v in config.flatten_dict.values():
            assert v == 1

        config1 = deepcopy(self.config)
        config1.update(config)
        assert config1 != self.config
        assert set(config1.paths) == set(self.config.paths)
        for v in config1.flatten_dict.values():
            assert v == 1

        config = deepcopy(self.config)
        fd = {f"not_exist{delimiter}not_exist{delimiter}not_exist": "not_exist"}
        try:
            config.update(fd)
            raise RuntimeError
        except Exception:
            pass
        config.update(fd, ignore_missing_path=True)
        assert "not_exist" in config
        assert f"not_exist{delimiter}not_exist" in config
        assert f"not_exist{delimiter}not_exist{delimiter}not_exist" in config
        assert config[f"not_exist{delimiter}not_exist{delimiter}not_exist"] == "not_exist"

        assert config != self.config

        try:
            config.update({f"task{delimiter}task": {"1": 1}})
            raise RuntimeError
        except Exception:
            pass

    def test_setitem(self, delimiter=";"):
        config = deepcopy(self.config)
        for path in config.flatten_dict.keys():
            config[path] = 1

        assert config != self.config
        assert set(config.paths) == set(self.config.paths)
        for v in config.flatten_dict.values():
            assert v == 1

        config = deepcopy(self.config)
        config[
            f"not_exist{delimiter}not_exist{delimiter}not_exist{delimiter}not_exist"
        ] = "not_exist"
        assert "not_exist" in config
        assert f"not_exist{delimiter}not_exist" in config
        assert f"not_exist{delimiter}not_exist{delimiter}not_exist" in config
        assert f"not_exist{delimiter}not_exist{delimiter}not_exist{delimiter}not_exist" in config
        assert (
            config[f"not_exist{delimiter}not_exist{delimiter}not_exist{delimiter}not_exist"]
            == "not_exist"
        )

        assert config != self.config

        try:
            config[f"task{delimiter}task"] = {1: 1}
            raise ValueError
        except Exception:
            pass

    def test_save(self, delimiter=";"):
        self.config.save("tmp.yaml")
        with open("tmp.yaml", "r") as f:
            d = yaml.safe_load(f)
        config = nconfig(d, delimiter=delimiter)
        assert config == self.config
        assert config._arg_specification == self.config._arg_specification
        os.system("rm -rf tmp.yaml")

    def test_delimiter(self):
        delimiter = "."
        self.nd = ndict(self.rawd, delimiter=delimiter)
        self.test_init(delimiter=delimiter)
        self.test_parse(delimiter=delimiter)
        with open("test/config_arg_spe.yaml", "r") as f:
            rawd_spe = yaml.safe_load(f)
        nd_spe = ndict(rawd_spe, delimiter=delimiter)
        self.config = nconfig(nd_spe, delimiter=delimiter)
        self.test_update(delimiter=delimiter)
        self.test_setitem(delimiter=delimiter)
        self.test_save(delimiter=delimiter)


testNDict()

testNConfig()



from __future__ import annotations

import json
from copy import deepcopy
from functools import reduce
from operator import getitem
from typing import Any, Optional, Union


class ndict:
    """Due to the internal mechanism, modifications of the raw_dict will not be applied on the
    object."""

    _missing_methods = ["ignore", "false", "error"]

    def __init__(
        self, dictionary: Optional[Union[ndict, dict]] = None, delimiter: str = ";"
    ) -> None:
        dictionary = dictionary or {}
        if isinstance(dictionary, ndict):
            self._d = dictionary.raw_dict
        elif isinstance(dictionary, dict):
            self._d = dictionary
        else:
            raise TypeError(f"Unexpected type {type(dictionary)}.")

        assert isinstance(delimiter, str), f"delimiter must be str, but recieved {type(delimiter)}"
        self.delimiter = delimiter
        self._update_flatten()

    @classmethod
    def from_flatten_dict(cls, flatten_dict: dict, delimiter=";") -> ndict:
        """Generate nested from flattened dictionary.
        The delimiter must be the same!
        """
        nd = cls({}, delimiter=delimiter)
        nd.update(flatten_dict)
        return nd

    @classmethod
    def from_list_of_dict(cls, ls: list, delimiter=";") -> ndict:
        """Generate nested from a list of dictionaries."""
        res = cls()
        for d in ls:
            res.update(d)
        return res

    def is_matched(self, query: dict, missing_method: str = "ignore") -> bool:
        """Test if the dictionary is the queried one.

        Syntax:
            1. Normal path-value pair: {path: value}.
            2. Query expression: {path: !QUERY [python expression returns boolean results]}. The
                expression can use d: the dictionary and path: the query path.

        Args:
            query (dict): query dictionary.
            missing_method (str): actions when the path is missing from the dictionary. Support:
            ['ignore', 'false', 'error'].
        Returns:
            (bool): if the dictionary is the queried one.
        """
        assert (
            missing_method in self._missing_methods
        ), f"Wrong missing method: {missing_method} / {self._missing_methods}."
        d = self
        for path, v in query.items():
            if path not in self:
                if missing_method == "ignore":
                    continue
                elif missing_method == "false":
                    return False
                elif missing_method == "error":
                    raise KeyError(f"Wrong path: {path}.")

            if isinstance(v, str) and v.startswith("!QUERY") and not eval(v[6:].strip()):
                return False
            if self[path] != v:
                return False
        return True

    ### internal manipulation ###
    def _update_flatten(self) -> None:
        flatten = {}
        paths = []

        def _flatten(pks, k, v, flatten, paths):
            ks = pks + [k]
            paths.append(self.delimiter.join(ks))
            if isinstance(v, dict):
                for nk, nv in v.items():
                    _flatten(ks, nk, nv, flatten, paths)
            else:
                flatten[paths[-1]] = v

        for k, v in self._d.items():
            _flatten([], k, v, flatten, paths)

        self._flatten_dict = flatten
        self._paths = paths

    ### getters ###
    def get(self, path, default=None):
        try:
            return self.__getitem__(path)
        except Exception:
            return default

    def compare_dict(self, other):
        assert isinstance(other, ndict)
        output = {}
        for path, v in self._flatten_dict.items():
            other_v = other.get(path, None)
            if other_v != v:
                output[path] = v, other_v

        for path, v in other._flatten_dict.items():
            self_v = self.get(path, None)
            if self_v != v:
                output[path] = self_v, v

        return output

    @property
    def raw_dict(self):
        return deepcopy(self._d)

    @property
    def flatten_dict(self):
        return deepcopy(self._flatten_dict)

    @property
    def flatten_dict_split(self):
        return deepcopy(
            [ndict.from_flatten_dict({p: v}).raw_dict for p, v in self.flatten_dict.items()]
        )

    @property
    def paths(self):
        return deepcopy(self._paths)

    @property
    def size(self):
        return len(self._flatten_dict)

    ### setters & updators ###
    def update(self, d, ignore_missing_path=False):
        if isinstance(d, dict):
            d = ndict(d, delimiter=self.delimiter)._flatten_dict
        elif isinstance(d, ndict):
            d = d._flatten_dict
        else:
            raise TypeError(f"Unexpected type {type(d)}")

        for path, v in d.items():
            if path not in self.paths and ignore_missing_path:
                continue
            self[path] = v

    ### iterations ###
    def items(self):
        return self._d.items()

    def keys(self):
        return self._d.keys()

    def values(self):
        return self._d.values()

    ### magics ###
    def __bool__(self):
        return bool(len(self._d))

    def __contains__(self, path: str) -> bool:
        return path in self._paths

    def __delitem__(self, path: str) -> None:
        if path in self._paths:
            if self.delimiter not in path:
                del self._d[path]
            else:
                path = path.split(self.delimiter)
                del reduce(getitem, path[:-1], self._d)[path[-1]]
            self._update_flatten()

    def __getitem__(self, path: str) -> Any:
        if path not in self._paths:
            raise KeyError(path)
        if self.delimiter not in path:
            return self._d[path]
        path = path.split(self.delimiter)
        return reduce(getitem, path, self._d)

    def __len__(self) -> int:
        return len(self._d)

    def __setitem__(self, path: str, value) -> None:
        if self.delimiter not in path:
            self._d[path] = value
        else:
            path = path.split(self.delimiter)
            v = self._d
            for node in path[:-1]:
                if node not in v:
                    v[node] = {}
                v = v[node]
            v[path[-1]] = value
        self._update_flatten()

    def __str__(self) -> str:
        return json.dumps(self._d, sort_keys=False, indent=2)

    def __eq__(self, other) -> bool:
        assert isinstance(other, ndict), f"Unexpected type {type(other)}"
        return self._flatten_dict == other._flatten_dict


NestedOrDict = Union[ndict, dict]



from __future__ import annotations

import argparse
from copy import deepcopy
from pathlib import Path
from typing import Union

import yaml

from .ndict import ndict, NestedOrDict


class CustomArgs:
    def __init__(
        self,
        flag: str = None,
        atype=None,
        path: str = None,
        nargs: str = None,
        choices: list = None,
        default=None,
        isbool: bool = False,
    ) -> None:
        self.flag = flag
        self.atype = atype
        self.path = path
        self.nargs = nargs
        self.choices = choices
        self.default = default
        self.isbool = isbool

    def __repr__(self):
        return f"{self.flag} {self.path}"


class nconfig(ndict):
    """Nested configuration based on NDict class.
    You are not supposed to modify the data outside this class!
    """

    _arg_key = "_ARGUMENT_SPECIFICATION"
    _ignore_key = "_IGNORE_IN_CLI"
    reserved_keys = [_arg_key, _ignore_key]

    # List is also allowed.
    allowable_types = [int, str, float, bool, type(None)]

    def __init__(self, config: NestedOrDict, delimiter: str = ";") -> None:
        # Configuration is supposed to be read-only. Therefore should not be modified in outside.
        config = deepcopy(config)
        if self._arg_key in config:
            self._arg_specification = config[self._arg_key]
            del config[self._arg_key]
        else:
            self._arg_specification = {}
        super().__init__(config, delimiter=delimiter)
        self._check_types()

    @property
    def raw_dict(self):
        return deepcopy(self._d)

    @property
    def flatten_dict(self):
        return deepcopy(self._flatten_dict)

    @property
    def flatten_dict_split(self):
        return deepcopy(
            [ndict.from_flatten_dict({p: v}).raw_dict for p, v in self.flatten_dict.items()]
        )

    @property
    def paths(self):
        return deepcopy(self._paths)

    def _get_custom_args(self) -> list[CustomArgs]:
        """Generate customized arguments."""
        custom_args = []
        for path, v in self._flatten_dict.items():
            spec = self._arg_specification.get(path, {})
            if spec == self._ignore_key:
                continue
            if isinstance(v, list):
                atype = type(v[0])
                nargs = "+"
            else:
                atype = type(v)
                nargs = None
            if isinstance(v, bool):
                isbool = True
                choices = [0, 1]
                atype = int
            else:
                isbool = False
                choices = None
            carg = {
                "flag": f"--{path.replace(self.delimiter, '__')}",
                "isbool": isbool,
                "atype": atype,
                "path": path,
                "nargs": nargs,
                "choices": choices,
                "default": v,
            }

            assert all(k != "atype" for k in spec.keys())
            carg.update(spec)
            custom_args.append(CustomArgs(**carg))

        return custom_args

    def add_to_argparse(self, parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        """Add arguments to parser.
        For a path "task;task", the argument is specified as "--task__task".
        The type is inferred from the configuration file.
        Use integer 1 and 0 for boolean values.
        For list, nargs are automatically configured.

        The flag, type, and other options can be customized. See README.md.
        Use "_IGNORE_IN_CLI" to omit the path when adding to parse.
        """
        custom_args = self._get_custom_args()
        for carg in custom_args:
            parser.add_argument(
                carg.flag,
                type=carg.atype,
                nargs=carg.nargs,
                choices=carg.choices,
                default=carg.default,
            )
        return parser

    def parse_update(self, parser, args) -> list:
        """Parse CLI arguments and update the configurations accordingly."""
        custom_args = self._get_custom_args()

        args, extra_args = parser.parse_known_args(args)
        for carg in custom_args:
            v = getattr(args, carg.flag.strip("-"))
            if v is not None:
                self[carg.path] = bool(v) if carg.isbool else v

        return extra_args

    ### getters ###
    @property
    def str_configs(self) -> nconfig:
        """Return NConfig object with every value in string."""
        d = deepcopy(self)
        for path, v in self._flatten_dict.items():
            d[path] = str(v)
        return d

    ### miscs ###
    def save(self, path: Union[Path, str]) -> None:
        """Save the configuraion. CLI specification will also be saved."""
        path = Path(path)
        path.parent.mkdir(exist_ok=True, parents=True)
        if self._arg_specification:
            d = {**self._d, self._arg_key: self._arg_specification}
        else:
            d = self._d
        with open(path, "w") as f:
            yaml.dump(d, f, indent=4, sort_keys=False)

    ### check ###
    def _check_type_v(self, v):
        """Check the data type for a single value. The allowaed dtype is stated
        in self.allowable_types.
        """
        return (
            all(any(isinstance(x, t) for t in self.allowable_types) for x in v)
            if isinstance(v, list)
            else any(isinstance(v, t) for t in self.allowable_types)
        )

    def _check_types(self):
        """Check the data type. Call _check_type_v on each entry."""
        for path, v in self.flatten_dict.items():
            assert self._check_type_v(
                v
            ), f"Unexpected config value type of {path} {type(v)} / {self.allowable_types}"

    ### magics ###
    def __setitem__(self, path: str, value: Union[str, int, float, bool]):
        """Same to NDict but do type checking."""
        self._check_type_v(value)
        super(nconfig, self).__setitem__(path, value)
